from django.core.exceptions import ValidationError
from arkindex.project.tests import FixtureTestCase
from arkindex.images.models import ImageServer
from arkindex.dataimport.config import ConfigFile, ImportType, VolumesImportFormat, ActsImportFormat


class TestConfigFile(FixtureTestCase):
    """
    Tests for the .arkindex.yml configuration file parser
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.sample_server = ImageServer.objects.create(display_name='sample', url='http://example.com/iiif/')

    def setUp(self):
        # A valid configuration; tests will edit/remove values to check for validations
        self.base_data = {
            "version": 1,
            "branches": ["master", "dev"],
            "volumes": {
                "format": "txt",
                "paths": ["volumes/*", "manifests/*"],
                "image_servers": {
                    "sample": "http://example.com/iiif/",
                },
                "lazy_checks": True,
                "autoconvert_https": True,
            },
            "acts": {
                "format": "csv",
                "paths": ["acts/*"],
            }
        }

    def test_base(self):
        """
        Test parsing a valid configuration file
        """
        cfg = ConfigFile(data=self.base_data)
        self.assertEqual(cfg.version, 1)
        self.assertListEqual(cfg.branches, ['master', 'dev'])
        self.assertCountEqual(cfg.imports, [ImportType.Volumes, ImportType.Acts])

        self.assertEqual(cfg.volumes_format, VolumesImportFormat.TXT)
        self.assertEqual(cfg.acts_format, ActsImportFormat.CSV)
        self.assertListEqual(cfg.volumes_paths, ['volumes/*', 'manifests/*'])
        self.assertListEqual(cfg.acts_paths, ['acts/*'])

        self.assertTrue(cfg.volumes_lazy_checks)
        self.assertTrue(cfg.volumes_autoconvert_https)
        self.assertEqual(cfg.volumes_image_servers, [self.sample_server])

    def test_version(self):
        """
        Test the ConfigFile handles versioning
        """
        self.base_data['version'] = 42  # Bad version number
        with self.assertRaisesRegex(ValidationError, 'version'):
            ConfigFile(data=self.base_data)

    def test_branches(self):
        """
        Test the ConfigFile handles branches
        """
        self.base_data['branches'] = ['this', 'is', 'a', {'bad': 'format'}]
        with self.assertRaisesRegex(ValidationError, 'branches'):
            ConfigFile(data=self.base_data)

    def test_no_imports(self):
        """
        Test the config parser requires at least one import
        """
        del self.base_data['volumes']
        del self.base_data['acts']
        with self.assertRaisesRegex(ValidationError, 'import types'):
            ConfigFile(data=self.base_data)

    def test_import_paths(self):
        """
        Test the config parser checks for valid paths
        """
        del self.base_data['volumes']['paths']
        with self.assertRaisesRegex(ValidationError, 'paths'):
            ConfigFile(data=self.base_data)

        self.base_data['volumes']['paths'] = ['bad', {'paths': 'format'}]
        with self.assertRaisesRegex(ValidationError, 'paths'):
            ConfigFile(data=self.base_data)

    def test_import_formats(self):
        """
        Test omitting import formats sets to default values
        """
        del self.base_data['volumes']['format']
        del self.base_data['acts']['format']
        cfg = ConfigFile(data=self.base_data)
        self.assertEqual(cfg.volumes_format, VolumesImportFormat.IIIF)
        self.assertEqual(cfg.acts_format, ActsImportFormat.CSV)
