from rest_framework import serializers
from arkindex_common.enums import EntityType, TranscriptionType
from arkindex.project.serializer_fields import EnumField
from arkindex.documents.date_parser import parse_date
from arkindex.documents.models import Element, ElementType, Entity
from arkindex.documents.serializers.light import CorpusLightSerializer
from arkindex.documents.serializers.elements import ElementSlimSerializer
from arkindex.documents.serializers.ml import TranscriptionSerializer
import math


class SearchQuerySerializer(serializers.Serializer):
    """
    Search parameters validation serializer in order to build a ES query

    date_lte is rounded to superior value using ES syntax (see round_date()) and served as date_lt (exactly lower than)
    Rounded value depend if month/day are provided (Ex: '<= march 1999' becomes '< april 1999')
    """
    q = serializers.CharField(source='query', default=None)
    score = serializers.FloatField(source='min_score', min_value=0.0, max_value=1.0, default=0.0)
    date_gte = serializers.CharField(default=None)
    date_lte = serializers.CharField(source='date_lt', default=None)
    element_type = EnumField(enum=ElementType, default=None)
    transcription_type = EnumField(enum=TranscriptionType, default=None)
    corpus = serializers.UUIDField(source='corpus_id', default=None)

    def parse_date(self, raw_date):
        date = parse_date(raw_date)
        if len(date) != 1:
            raise serializers.ValidationError('Could not parse Date')
        return date[0]

    def validate_score(self, value):
        if not isinstance(value, float) or math.isinf(value) or math.isnan(value):
            raise serializers.ValidationError('Score should be a valid decimal number between 0 and 1')
        return value

    def validate_date_gte(self, value):
        if not value:
            return
        return self.parse_date(value).es_str()

    def validate_date_lte(self, value):
        if not value:
            return
        return self.parse_date(value).es_round_up()

    def validate(self, data):
        data = super().validate(data)
        gte = data['date_gte']
        lt = data['date_lt']
        if gte and lt and gte > lt:
            raise serializers.ValidationError("Upper date must be greater than lower date")
        if not (data['query'] or data['date_lt'] or data['date_gte']):
            raise serializers.ValidationError("At least query terms or dates must be provided")
        return data


class EntitySearchQuerySerializer(serializers.Serializer):
    """
    Search parameters validation serializer in order to build a ES query for entity
    """
    q = serializers.CharField(source='query')
    type = EnumField(enum=EntityType, default=None)
    corpus = serializers.UUIDField(source='corpus_id', default=None)


class IIIFSearchQuerySerializer(serializers.Serializer):
    """
    Search parameters for IIIF transcription search
    See https://iiif.io/api/search/1.0/#request-1
    """
    q = serializers.CharField(source='query')


class ElementSearchResultSerializer(serializers.ModelSerializer):
    """
    A page search result with nested transcriptions
    """
    type = EnumField(ElementType)
    transcriptions = TranscriptionSerializer(many=True, source='transcriptions_results')
    total_transcriptions = serializers.IntegerField()
    parent_paths = serializers.ListField(
        child=serializers.ListField(
            child=ElementSlimSerializer()
        ),
    )
    corpus = CorpusLightSerializer()
    date_from = serializers.CharField()
    date_to = serializers.CharField()

    class Meta:
        model = Element
        fields = (
            'id',
            'name',
            'type',
            'transcriptions',
            'total_transcriptions',
            'parent_paths',
            'corpus',
            'date_from',
            'date_to',
        )


class EntitySearchResultSerializer(serializers.ModelSerializer):
    """
    Serializes an entity
    """
    type = EnumField(EntityType)

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas'
        )
