from itertools import groupby
from arkindex_common.tei import TeiElement, TeiParser as BaseTeiParser
from arkindex_common.enums import MetaType
from arkindex.project.tools import find_closest
from arkindex.documents.indexer import Indexer
from arkindex.documents.models import Element, ElementType, Entity, MetaData, DataSource, MLToolType
import logging


logger = logging.getLogger(__name__)


class TeiParser(BaseTeiParser):

    def __init__(self, path, corpus, revision):
        super().__init__(path)
        self.db_corpus = corpus
        self.revision = revision

    def match_database(self):
        '''
        Link TEI elements with DB elements
        Outputs a list of tuples (db_element, tei_element)
        '''
        volumes = self.db_corpus.elements.filter(type=ElementType.Volume)
        assert volumes.count() > 0, 'No volumes available'

        # Match volumes
        out = []
        for tei in self.corpus.tei:
            if not tei.witness:
                logger.warning('No witness in {}'.format(str(tei)))
                continue

            tei_name = tei.witness.id or tei.witness.repository_id
            volume = find_closest(tei_name, volumes)
            if not volume:
                logger.warning('No matching volume for {}'.format(tei))
                continue

            out.append((volume, tei))
            logger.info('Matched {} with {}'.format(volume, tei))

            # Load volume acts
            volume_acts = Element.objects.get_descending(volume.id, type=ElementType.Act)
            if not volume_acts.exists():
                logger.warning('No acts in DB for {}'.format(volume), volume)
                continue

            # Match acts
            for text in tei.texts:
                if text.witness is None:
                    logger.warning('No witness on text, skipping.')
                    continue

                act = Element.objects.filter(
                    id__in=volume_acts,
                    type=ElementType.Act,
                    metadatas__name='number',
                    metadatas__value=text.witness.id,
                ).first()
                if act:
                    out.append((act, text))
                    logger.info('Matched {} with {}'.format(act, text))
                else:
                    logger.warning('No match for {}'.format(text))

        return out

    def save(self, tei_elt, db_elt):
        """
        Save or update all metadatas contained in TEI element to corresponding database element
        Return a list of saved Metadatas
        """
        assert isinstance(tei_elt, TeiElement)
        assert isinstance(db_elt, Element)

        # Filter to remove empty values
        parsed_metadatas = list(filter(lambda meta: meta.value, tei_elt.build_metadata()))

        # Group by name
        def grouper(metadata):
            return metadata.name

        parsed_metadatas = groupby(sorted(parsed_metadatas, key=grouper), key=grouper)
        db_metadatas = []

        # Create/update
        for name, metadatas in parsed_metadatas:
            for index, xmlmeta in enumerate(metadatas):
                # Handle entities
                entity = None
                if xmlmeta.type == MetaType.Entity and xmlmeta.entity_type is not None:
                    source, _ = DataSource.objects.get_or_create(
                        type=MLToolType.NER,
                        slug='TEI',
                        internal=False
                    )
                    entity, _ = Entity.objects.get_or_create(
                        corpus=db_elt.corpus,
                        name=xmlmeta.value,
                        type=xmlmeta.entity_type,
                        source=source
                    )

                db_meta, created = db_elt.metadatas.get_or_create(
                    name=name,
                    index=index,
                    defaults={
                        'type': xmlmeta.type,
                        'value': xmlmeta.value,
                        'revision': self.revision,
                        'entity': entity,
                    },
                )
                db_metadatas.append(db_meta)
                if created:
                    continue
                if (db_meta.type, db_meta.value) == (xmlmeta.type, xmlmeta.value) \
                        and db_meta.revision \
                        and (not db_meta.entity or db_meta.entity.name == xmlmeta.value):
                    # Nothing to update and revision is set
                    continue
                db_meta.type, db_meta.value, db_meta.entity, db_meta.revision = \
                    xmlmeta.type, xmlmeta.value, entity, self.revision
                db_meta.save()

        # Remove deleted metadatas
        db_elt.metadatas.exclude(
            id__in=[dbm.id for dbm in db_metadatas],
        ).delete()
        return db_metadatas

    def run(self):
        matches = self.match_database()
        logger.info('Found {} metadatas matching DB'.format(len(matches)))
        db_metadatas = []
        for db_elt, tei_elt in matches:
            db_metadatas += self.save(tei_elt, db_elt)
        self.index(db_metadatas)

    def index(self, db_metadatas):
        """
        Create or Update ElasticSearch index for indexed elements
        """
        md_ids = (md.id for md in db_metadatas)
        date_metadatas = MetaData.objects.filter(id__in=md_ids, type=MetaType.Date)

        acts = Element.objects \
            .filter(metadatas__in=date_metadatas, type=ElementType.Act) \
            .prefetch_related('metadatas')
        if acts.exists():
            Indexer().run_index(acts)

        pages = Element.objects \
            .filter(metadatas__in=date_metadatas, type=ElementType.Page) \
            .prefetch_related('transcriptions')
        if pages.exists():
            Indexer().run_index(pages)
