from django.core.management import call_command
from django.db.models.signals import pre_delete
from arkindex_common.enums import TranscriptionType, MetaType, DataImportMode
from arkindex_common.ml_tool import MLToolType
from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import Corpus, Element, ElementType, DataSource, MLClass
from arkindex.dataimport.models import EventType


class TestDeleteCorpus(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.img1 = cls.imgsrv.images.get(path='img1')
        cls.zone = cls.img1.zones.create(polygon=[(0, 0), (10, 0), (10, 10), (0, 10), (0, 0)])
        cls.corpus2 = Corpus.objects.create(name='Other corpus')
        cls.corpus2.create_default_himanis_types()
        cls.repo = cls.corpus2.repos.create(
            url='http://lol.git',
            hook_token='h00k',
            provider_name='gitlab',
        )
        cls.rev = cls.repo.revisions.create(
            hash='42',
            ref='master',
            message='oh',
            author='me',
        )
        cls.di = cls.rev.dataimports.create(
            creator=cls.user,
            corpus=cls.corpus2,
            mode=DataImportMode.Repository,
        )
        cls.df = cls.di.files.create(
            name='a.txt',
            size=1234,
            hash='c0ffee',
            content_type='text/plain',
            corpus=cls.corpus2,
        )
        cls.vol = cls.corpus2.elements.create(
            type=ElementType.Volume,
            new_type=cls.corpus2.himanis_volume,
            name='A volume',
        )
        cls.vol.metadatas.create(
            type=MetaType.Text,
            name='meta',
            value='data',
        )
        cls.vol.events.create(
            revision=cls.rev,
            type=EventType.Addition,
        )
        cls.page = Element.objects.create(
            corpus=cls.corpus2,
            type=ElementType.Page,
            new_type=cls.corpus2.himanis_page,
            name='A page',
        )
        cls.page.classifications.create(
            source=DataSource.objects.create(
                type=MLToolType.Classifier,
                slug='classeur',
                revision='Early Access',
                internal=False,
            ),
            ml_class=MLClass.objects.create(
                name='klass',
                corpus=cls.corpus2,
            ),
            confidence=0.5,
        )
        cls.page.transcriptions.create(
            source=DataSource.objects.create(
                type=MLToolType.Recognizer,
                slug='reco',
                revision='-1',
                internal=False,
            ),
            zone=cls.zone,
            type=TranscriptionType.Word,
            text='hi',
            score=0.75,
        )
        cls.page.add_parent(cls.vol)

    def test_dry_run(self):
        self.assertEqual(Corpus.objects.count(), 2)
        self.assertEqual(self.corpus.elements.count(), 19)
        self.assertEqual(self.corpus2.elements.count(), 2)
        self.assertEqual(self.imgsrv.images.count(), 6)

        call_command(
            'delete_corpus',
            self.corpus,
            dry_run=True,
        )
        self.corpus.refresh_from_db()
        self.corpus2.refresh_from_db()
        self.imgsrv.refresh_from_db()

        self.assertEqual(Corpus.objects.count(), 2)
        self.assertEqual(self.corpus.elements.count(), 19)
        self.assertEqual(self.corpus2.elements.count(), 2)
        self.assertEqual(self.imgsrv.images.count(), 6)

    def test_run(self):
        receivers = pre_delete.receivers

        call_command(
            'delete_corpus',
            self.corpus,
        )

        # Ensure the command restores the signal receivers
        self.assertEqual(pre_delete.receivers, receivers)

        self.assertEqual(Corpus.objects.count(), 1)
        with self.assertRaises(Corpus.DoesNotExist):
            self.corpus.refresh_from_db()
        self.corpus2.refresh_from_db()
        self.assertEqual(self.corpus2.types.count(), 4)
        self.assertEqual(self.corpus2.elements.count(), 2)
        self.assertEqual(Element.objects.count(), 2)

        self.repo.refresh_from_db()
        self.rev.refresh_from_db()
        self.di.refresh_from_db()
        self.df.refresh_from_db()
        self.vol.refresh_from_db()
        self.page.refresh_from_db()

        self.assertTrue(self.repo.revisions.filter(id=self.rev.id).exists())
        self.assertEqual(self.di.revision, self.rev)
        self.assertEqual(self.di.files.get(), self.df)
        self.assertTrue(Element.objects.get_descending(self.vol.id).filter(id=self.page.id).exists())

        md = self.vol.metadatas.get()
        self.assertEqual(md.name, 'meta')
        self.assertEqual(md.value, 'data')

        event = self.vol.events.get()
        self.assertEqual(event.type, EventType.Addition)
        self.assertEqual(event.revision, self.rev)

        cl = self.page.classifications.get()
        self.assertEqual(cl.source.slug, 'classeur')
        self.assertEqual(cl.ml_class.name, 'klass')
        self.assertEqual(cl.confidence, 0.5)

        ts = self.page.transcriptions.get()
        self.assertEqual(ts.source.slug, 'reco')
        self.assertEqual(ts.zone, self.zone)
        self.assertEqual(ts.type, TranscriptionType.Word)
        self.assertEqual(ts.text, 'hi')
        self.assertEqual(ts.score, 0.75)
