from django.core.management import call_command
from django.core.management.base import CommandError
from unittest.mock import patch, call
from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import Corpus, ElementType


class TestGenerateThumbnailsCommand(FixtureTestCase):
    """
    Test the generate_thumbnails command.
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol1 = cls.corpus.elements.get(type=ElementType.Volume, name="Volume 1")
        cls.vol2 = cls.corpus.elements.get(type=ElementType.Volume, name="Volume 2")
        corpus2 = Corpus.objects.create(name='Other corpus')
        corpus2.create_default_himanis_types()
        cls.vol3 = corpus2.elements.create(type=ElementType.Volume, new_type=corpus2.himanis_volume, name='Volume 3')
        cls.thumb_patch = patch('arkindex.images.models.Thumbnail')

    def setUp(self):
        super().setUp()
        self.thumb_mock = self.thumb_patch.start()

    def tearDown(self):
        super().tearDown()
        self.thumb_patch.stop()

    def test_start_corpus(self):
        """
        Test generate_thumbnails runs generation for each volume of a corpus
        """
        call_command(
            'generate_thumbnails',
            corpus=self.corpus,
        )
        self.assertCountEqual(self.thumb_mock.call_args_list, [
            call(self.vol1),
            call(self.vol2),
        ])
        self.assertEqual(self.thumb_mock().create.call_count, 2)

    def test_start_element(self):
        """
        Test generate_thumbnails runs generation for an element
        """
        call_command(
            'generate_thumbnails',
            corpus=self.corpus,
            element=self.vol1,
        )
        self.assertCountEqual(self.thumb_mock.call_args_list, [
            call(self.vol1),
        ])
        self.assertEqual(self.thumb_mock().create.call_count, 1)

    @patch('arkindex.documents.models.Element.generate_thumbnail')
    def test_force(self, gen_mock):
        """
        Test generate_thumbnails passes the --force argument to tasks
        """
        call_command(
            'generate_thumbnails',
            corpus=self.corpus,
            element=self.vol1,
            force=True,
        )
        self.assertCountEqual(gen_mock.call_args_list, [
            call(force=True),
        ])

    def test_all(self):
        """
        Test generate_thumbnails picks all volumes when using --all
        """
        call_command(
            'generate_thumbnails',
            all=True,
        )
        self.assertCountEqual(self.thumb_mock.call_args_list, [
            call(self.vol1),
            call(self.vol2),
            call(self.vol3),
        ])
        self.assertEqual(self.thumb_mock().create.call_count, 3)

    def test_all_xor_corpus(self):
        """
        Test generate_thumbnails does not allow --all and --corpus simultaneously
        """
        with self.assertRaisesRegex(CommandError, r'--all.+--corpus'):
            call_command(
                'generate_thumbnails',
                all=True,
                corpus=self.corpus,
            )

    def test_corpus_required(self):
        """
        Test generate_thumbnails requires either --all or --corpus
        """
        with self.assertRaisesRegex(CommandError, '--corpus'):
            call_command('generate_thumbnails')

    def test_element_in_corpus(self):
        """
        Test generate_thumbnails requires --element to be inside --corpus
        """
        with self.assertRaisesRegex(CommandError, 'not in corpus'):
            call_command(
                'generate_thumbnails',
                corpus=self.corpus,
                element=self.vol3,
            )
