from django.contrib.auth.models import AnonymousUser
from django.db.models import Model, ProtectedError
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.project.tests import FixtureAPITestCase


class TestCorpus(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.anon = AnonymousUser()

        cls.corpus_public = Corpus.objects.get(name='Unit Tests', public=True)
        cls.corpus_private = Corpus.objects.create(name='B Private')
        cls.user.corpus_right.create(corpus=cls.corpus_private, user=cls.user, can_write=True)
        cls.corpus_hidden = Corpus.objects.create(name='C Hidden')

    def test_anon(self):
        # An anonymous user has only access to public
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()

        self.assertEqual(len(data), 1)
        # Ignore ordering on types
        self.assertCountEqual(data[0].pop('types'), [
            {
                'slug': 'volume',
                'display_name': 'Volume',
                'folder': True,
                'hidden': False
            },
            {
                'slug': 'page',
                'display_name': 'Page',
                'folder': False,
                'hidden': False
            },
            {
                'slug': 'act',
                'display_name': 'Act',
                'folder': False,
                'hidden': False
            },
            {
                'slug': 'surface',
                'display_name': 'Surface',
                'folder': False,
                'hidden': True
            }
        ])

        self.assertListEqual(
            data,
            [
                {
                    'id': str(self.corpus_public.id),
                    'rights': ['read'],
                    'public': True,
                    'name': 'Unit Tests',
                    'description': '',
                }
            ]
        )

    def test_user(self):
        # A normal user has access to public + its private
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()

        self.assertEqual(len(data), 2)
        self.assertListEqual(data[0].pop('types'), [])
        self.assertCountEqual(data[1].pop('types'), [
            {
                'slug': 'volume',
                'display_name': 'Volume',
                'folder': True,
                'hidden': False
            },
            {
                'slug': 'page',
                'display_name': 'Page',
                'folder': False,
                'hidden': False
            },
            {
                'slug': 'act',
                'display_name': 'Act',
                'folder': False,
                'hidden': False
            },
            {
                'slug': 'surface',
                'display_name': 'Surface',
                'folder': False,
                'hidden': True
            }
        ])

        self.assertListEqual(
            data,
            [
                {
                    'id': str(self.corpus_private.id),
                    'rights': ['read', 'write'],
                    'public': False,
                    'name': 'B Private',
                    'description': '',
                },
                {
                    'id': str(self.corpus_public.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': True,
                    'name': 'Unit Tests',
                    'description': '',
                }
            ]
        )

    def test_admin(self):
        # An admin has access to everything
        self.client.force_login(self.superuser)
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()

        self.assertEqual(len(data), 3)
        self.assertListEqual(data[0].pop('types'), [])
        self.assertListEqual(data[1].pop('types'), [])
        self.assertCountEqual(data[2].pop('types'), [
            {
                'slug': 'volume',
                'display_name': 'Volume',
                'folder': True,
                'hidden': False
            },
            {
                'slug': 'page',
                'display_name': 'Page',
                'folder': False,
                'hidden': False
            },
            {
                'slug': 'act',
                'display_name': 'Act',
                'folder': False,
                'hidden': False
            },
            {
                'slug': 'surface',
                'display_name': 'Surface',
                'folder': False,
                'hidden': True
            }
        ])

        self.assertListEqual(
            data,
            [
                {
                    'id': str(self.corpus_private.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': False,
                    'name': 'B Private',
                    'description': '',
                },
                {
                    'id': str(self.corpus_hidden.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': False,
                    'name': 'C Hidden',
                    'description': '',
                },
                {
                    'id': str(self.corpus_public.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': True,
                    'name': 'Unit Tests',
                    'description': '',
                }
            ]
        )

    def test_mixin(self):
        vol1 = Element.objects.get(name="Volume 1")
        vol2 = Element.objects.get(name="Volume 2")

        response = self.client.get('{}?corpus={}&type=volume'.format(
            reverse('api:elements'),
            str(self.corpus_private.id),
        ))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        response = self.client.get('{}?corpus={}&type=volume'.format(
            reverse('api:elements'),
            str(self.corpus_public.id),
        ))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data['results']), 2)
        self.assertCountEqual(
            [r['id'] for r in data['results']],
            [str(vol1.id), str(vol2.id)],
        )

    def test_create(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus'), {
            'name': 'New Corpus',
            'description': 'Some description',
            'public': False,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        corpus = Corpus.objects.get(name='New Corpus', description='Some description', public=False)
        right = corpus.corpus_right.get(user=self.user)
        self.assertTrue(right.can_write)
        self.assertTrue(right.can_admin)
        self.assertEqual(corpus.types.count(), 4)
        self.assertCountEqual(
            corpus.types.values('slug', 'display_name', 'folder', 'hidden'),
            [
                {
                    'slug': 'volume',
                    'display_name': 'Volume',
                    'folder': True,
                    'hidden': False
                },
                {
                    'slug': 'page',
                    'display_name': 'Page',
                    'folder': False,
                    'hidden': False,
                },
                {
                    'slug': 'act',
                    'display_name': 'Act',
                    'folder': False,
                    'hidden': False,
                },
                {
                    'slug': 'surface',
                    'display_name': 'Surface',
                    'folder': False,
                    'hidden': True,
                },
            ]
        )

    def test_create_requires_login(self):
        response = self.client.post(reverse('api:corpus'), {
            'name': 'New Corpus',
            'description': 'Some description',
            'public': False,
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_public(self):
        response = self.client.get(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_public.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertCountEqual(data.pop('types'), [
            {
                'slug': 'volume',
                'display_name': 'Volume',
                'folder': True,
                'hidden': False
            },
            {
                'slug': 'page',
                'display_name': 'Page',
                'folder': False,
                'hidden': False,
            },
            {
                'slug': 'act',
                'display_name': 'Act',
                'folder': False,
                'hidden': False,
            },
            {
                'slug': 'surface',
                'display_name': 'Surface',
                'folder': False,
                'hidden': True,
            },
        ])
        self.assertDictEqual(data, {
            'id': str(self.corpus_public.id),
            'name': self.corpus_public.name,
            'description': self.corpus_public.description,
            'public': True,
            'rights': ['read'],
        })

    def test_retrieve(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.corpus_private.id),
            'name': self.corpus_private.name,
            'description': self.corpus_private.description,
            'public': False,
            'rights': ['read', 'write'],
            'types': [],
        })

    def test_retrieve_requires_login(self):
        response = self.client.get(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update(self):
        self.client.force_login(self.user)
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'name': 'new name',
            'description': 'new description',
        })
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.corpus_private.refresh_from_db()
        self.assertEqual(self.corpus_private.name, 'new name')
        self.assertEqual(self.corpus_private.description, 'new description')

    def test_update_requires_login(self):
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'name': 'new name',
            'description': 'new description',
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(Corpus.objects.filter(id=self.corpus_private.id).exists())

    def test_delete_requires_login(self):
        response = self.client.delete(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_elements_first(self):
        """
        Ensure corpus deletion deletes elements first to prevent a ProtectedError
        """
        self.assertEqual(Corpus.objects.count(), 3)
        self.assertEqual(Element.objects.count(), 19)

        # Use Django's default deletion method and expect a ProtectedError
        with self.assertRaises(ProtectedError):
            Model.delete(self.corpus)

        self.assertEqual(Corpus.objects.count(), 3)
        self.assertEqual(Element.objects.count(), 19)

        # Use the corpus deletion method and it works!
        self.corpus.delete()
        self.assertEqual(Corpus.objects.count(), 2)
        self.assertEqual(Element.objects.count(), 0)
