from arkindex_common.enums import MetaType, EntityType
from arkindex_common.tei import Text, NAMESPACES
from arkindex.documents.tei import TeiParser
from arkindex.project.tests import FixtureTestCase
import os.path

FIXTURES = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'tei_samples',
)


class TestTeiElement(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = cls.user.credentials.get().repos.get()

    def test_apply_xslt(self):
        te = Text(os.path.join(FIXTURES, 'arguments.xml'), NAMESPACES)
        self.assertMultiLineEqual(
            # Remove indentations
            "".join(line.strip() for line in te.argument.splitlines()),
            '<div xmlns:tei="http://www.tei-c.org/ns/1.0" class="argument">'
            '<p>Argument</p>'
            '<p><em>Italique</em></p>'
            '<p><sup>Supérieur</sup></p>'
            '<p><sup>Supérieur</sup></p>'
            '<p><blockquote>Citation</blockquote></p>'
            '</div>',
        )

    def test_save_update(self):
        """
        Check saving metadata for a TeiElement updates existing data
        """
        metadata_list = [
            {
                'name': 'folio',
                'index': 0,
                'type': MetaType.Text,
                'value': '1r',
                'entity__name': None,
                'entity__type': None,
            },
            {
                'name': 'date',
                'index': 0,
                'type': MetaType.Date,
                'value': '1337-01',
                'entity__name': None,
                'entity__type': None,
            },
            {
                'name': 'id',
                'index': 0,
                'type': MetaType.Text,
                'value': 'test_update',
                'entity__name': None,
                'entity__type': None,
            },
            {
                'name': 'language',
                'index': 0,
                'type': MetaType.Text,
                'value': 'lat.',
                'entity__name': None,
                'entity__type': None,
            },
            {
                'name': 'name',
                'index': 0,
                'type': MetaType.Text,
                'value': '1',
                'entity__name': None,
                'entity__type': None,
            },
            {
                'name': 'location',
                'index': 0,
                'type': MetaType.Entity,
                'value': 'Mars',
                'entity__name': 'Mars',
                'entity__type': EntityType.Location,
            },
            {
                'name': 'person',
                'index': 1,
                'type': MetaType.Entity,
                'value': 'The Chosen One',
                'entity__name': 'The Chosen One',
                'entity__type': EntityType.Person,
            },
        ]

        rev1 = self.repo.revisions.get()
        rev2 = self.repo.revisions.create(hash='43', ref='ref/heads/master', message='b', author='me')

        xml_path = os.path.join(FIXTURES, 'arguments.xml')
        parser = TeiParser(xml_path, corpus=self.corpus, revision=rev1)

        act = self.corpus.elements.get(name='Act 1')
        te_before = Text(os.path.join(FIXTURES, 'update_before.xml'), NAMESPACES)
        te_after = Text(os.path.join(FIXTURES, 'update_after.xml'), NAMESPACES)

        parser.save(te_before, act)
        self.assertFalse(act.metadatas.exclude(revision=rev1).exists())
        self.assertCountEqual(
            act.metadatas
               .exclude(name='summary')
               .values('name', 'index', 'type', 'value', 'entity__name', 'entity__type'),
            metadata_list + [
                    {
                        'name': 'person',
                        'index': 0,
                        'type': MetaType.Entity,
                        'value': 'Someone',
                        'entity__name': 'Someone',
                        'entity__type': EntityType.Person,
                    },
                    {
                        'name': 'place',
                        'index': 0,
                        'type': MetaType.Entity,
                        'value': 'Somewhere',
                        'entity__name': 'Somewhere',
                        'entity__type': EntityType.Location,
                    },
                ]
        )

        # Create subject, update person and delete place, leave location untouched
        parser.revision = rev2
        parser.save(te_after, act)
        self.assertCountEqual(
            act.metadatas
               .exclude(name='summary')
               .values('name', 'index', 'type', 'value', 'entity__name', 'entity__type'),
            metadata_list + [
                    {
                        'name': 'person',
                        'index': 0,
                        'type': MetaType.Entity,
                        'value': 'Someone else',
                        'entity__name': 'Someone else',
                        'entity__type': EntityType.Person,
                    },
                    {
                        'name': 'subject',
                        'index': 0,
                        'type': MetaType.Entity,
                        'value': 'Something',
                        'entity__name': 'Something',
                        'entity__type': EntityType.Subject,
                    },
                ]
        )
