from collections import namedtuple, defaultdict
from django.db import connection
from django.conf import settings
from pathlib import Path
from arkindex_common.tools import Timer
from arkindex_common.enums import TranscriptionType
from arkindex.project.polygon import Polygon
from arkindex.documents.indexer import Indexer
from arkindex.documents.models import Element, ElementType, Transcription, DataSource, Corpus
from arkindex.images.models import Image
import csv
import io
import os
import re
import gzip
import logging
import uuid
import multiprocessing as mp

REGEX_INDEX = re.compile(
    br'^(?:line_(\d+) )?(.+) \d+ ([\de\-\.]+) (\d+) (\d+) (\d+) (\d+)')

logger = logging.getLogger(__name__)

# Link a transcription data with a Polygon
# This is hashable (box is hashable)
TrPolygon = namedtuple('TrPolygon', 'image_id, element_id, source_id, type, polygon, text, score')


def extract_transcriptions(image, index_path):
    """
    Extract transcription data from index files
    One gzipped index file per image
    Format: #line word n/a score-[0,1] x1 y1 x2 y2
    """
    assert isinstance(image, Image)
    assert os.path.exists(index_path), \
        'Missing index {}'.format(index_path)
    assert index_path.endswith('.idx.gz')

    # Parse lines content
    lines = []
    with gzip.open(index_path) as f:
        for line in f.readlines():
            if line.decode('utf-8').startswith('#'):
                continue
            index = REGEX_INDEX.match(line)
            if index is None:
                logger.warning('Index parsing failed : {}'.format(line))
                continue

            # Build zone
            data = index.groups()
            x, y = int(data[3]), int(data[4])
            lines.append({
                'x': x,
                'y': y,
                'width': int(data[5]) - x,
                'height': int(data[6]) - y,
                'text': data[1].decode('utf-8'),
                'line': int(data[0] or 0),
                'score': float(data[2]),
            })

    logger.debug('Parsed {} lines'.format(len(lines)))
    return lines


def build_transcriptions(image, parent, source, items):
    """
    Build needed TrPolygon namedtuples from parsed index files
    """
    assert isinstance(parent, Element)

    # Build all TrPolygon from items
    required = {
        TrPolygon(
            image.id,
            parent.id,
            source.id,
            i.get('type', TranscriptionType.Word),
            Polygon.from_dict(i),
            i['text'],
            float(i['score']),
        )
        for i in items
    }

    # Build all TrPolygon from existing
    existing = {
        TrPolygon(
            tr.zone.image_id,
            tr.element_id,
            tr.source_id,
            tr.type,
            tr.zone.polygon,
            tr.text,
            tr.score,
        )
        for tr in parent.transcriptions.all().prefetch_related('zone')
    }

    # Calc needed TrPolygon to build
    return list(required.difference(existing))


def save_transcriptions(*tr_polygons, delimiter='\t', quotechar='"'):
    """
    Save TrPolygons in the database
    Uses PostgreSQL's COPY, as it is the fastest way to import, by generating CSV data in file-like objects
    """

    assert len(tr_polygons) > 0, \
        'No polygons, so no transcriptions to save.'

    import datetime
    now = datetime.datetime.utcnow().isoformat()

    def new_transcription(tr_polygon):
        zone_id = uuid.uuid4()
        return (
            (
                uuid.uuid4(),
                tr_polygon.element_id,
                tr_polygon.source_id,
                zone_id,
                tr_polygon.type.value,
                tr_polygon.text,
                tr_polygon.score,
            ),
            (
                zone_id,
                tr_polygon.image_id,
                tr_polygon.polygon.serialize(as_string=True),
                now,
                now,
            ),
        )

    def _list_to_csv(rows):
        """
        Convert a two dimensional iterable to a PostgreSQL COPY-compatible file-like object
        """
        out = io.StringIO()
        writer = csv.writer(out, dialect='unix', delimiter=delimiter, quotechar=quotechar, quoting=csv.QUOTE_MINIMAL)
        writer.writerows(rows)
        out.seek(0)
        return out

    # Prepare transcriptions & zones rows
    transcriptions, zones = zip(*map(new_transcription, tr_polygons))

    # Convert to CSV and run COPY via psycopg2
    with connection.cursor() as cursor:
        cursor.copy_from(
            _list_to_csv(zones),
            'images_zone',
            columns=('id', 'image_id', 'polygon', 'created', 'updated'),
            sep=delimiter,
        )
        cursor.copy_from(
            _list_to_csv(transcriptions),
            'documents_transcription',
            columns=('id', 'element_id', 'source_id', 'zone_id', 'type', 'text', 'score'),
            sep=delimiter,
        )

    return (transcriptions, zones)


def index_transcriptions(transcriptions):
    '''
    Index in ElasticSearch new transcriptions built above
    '''
    assert all(map(lambda t: isinstance(t, tuple), transcriptions))

    # Index transcriptions directly (IIIF search)
    indexer = Indexer()
    indexer.run_index(
        Transcription.objects.filter(
            id__in=[t[0] for t in transcriptions]
        ),
    )

    # Index transcriptions in pages
    pages = Element.objects.filter(id__in=[t[1] for t in transcriptions], type=ElementType.Page)
    if pages.exists():
        indexer.run_index(pages)


class IndexImporter(object):
    """Import index files (.idx.gz) as transcriptions."""

    def __init__(self, path, corpus, source):
        assert isinstance(corpus, Corpus)
        assert isinstance(source, DataSource)
        self.path = Path(path)
        assert self.path.is_dir()
        self.corpus = corpus
        self.source = source
        self.volumes = {}

    def find_volume(self, name):
        """
        Get an Element instance from a volume name by case-insensitive match in the corpus
        """
        if name in self.volumes:
            return self.volumes[name]
        try:
            self.volumes[name] = self.corpus.elements.get(type=ElementType.Volume, name__iexact=name)
            logger.debug('Matched volume "{}" for volume name "{}"'.format(self.volumes[name], name))
            return self.volumes[name]
        except Element.DoesNotExist:
            logger.warning('No matching volume found for "{}"'.format(name))
        except Element.MultipleObjectsReturned:
            logger.warning('Multiple matching volumes found for "{}"'.format(name))

    def find_page(self, pages, path):
        """
        In a given list of pages, find a page that matches a given filename
        """
        name = path.stem
        if name.endswith('.idx'):
            name = name[:-4]
        try:
            if name.startswith('f') and name[1:].isdigit():
                # Special case for Gallica folios in the f[0-9]+ form
                # Do not use 'in' here because 'f1.idx.gz' could return the 'f100' page
                return next(p for p in pages if p.zone.image.path.rpartition('/')[2] == name)

            return next(p for p in pages if name in p.zone.image.path.rpartition('/')[2])
        except StopIteration:
            logger.warning("No matching page for folio '{}' from file '{}'".format(name, str(path)))

    def list_index_pages(self):
        """
        Get (pages list, path) tuples for index files
        """
        # Get paths grouped by volumes
        volume_paths = defaultdict(list)
        for path in self.path.glob('**/*.idx.gz'):
            rel_path = path.relative_to(self.path)
            if len(rel_path.parts) <= 1:  # File is in root folder
                logger.warning('File {} is not in a subfolder'.format(str(rel_path)))
                continue
            volume = self.find_volume(rel_path.parts[0])
            if not volume:
                continue
            volume_paths[volume].append(path)

        # Load pages once for each volume, then yield pages and paths
        for volume, paths in volume_paths.items():
            pages = list(
                Element.objects
                       .get_descending(volume.id)
                       .filter(type=ElementType.Page)
                       .prefetch_related('zone', 'zone__image'))
            for path in paths:
                yield pages, path

    def run(self):
        from django import db
        db.connections.close_all()

        transcriptions = mp.JoinableQueue(settings.TRANSCRIPTIONS_IMPORT_QUEUE_SIZE)
        build = mp.Process(target=self.build_transcriptions, args=(transcriptions, ))
        save = mp.Process(target=self.save_transcriptions, args=(transcriptions, ))

        build.start()
        save.start()

        # Run until all files are parsed and the queue is filled
        build.join()

        # Wait for the entire queue to be processed
        transcriptions.join()

        logger.info('Import complete')
        save.terminate()

    def build_transcriptions(self, queue):
        '''
        Build needed transcriptions in a process
        '''
        for pages, index_path in self.list_index_pages():
            logger.info("Parsing index file {}".format(str(index_path)))
            page = self.find_page(pages, index_path)
            if not page:
                continue
            image = page.zone.image

            with Timer() as t:
                lines = extract_transcriptions(image, page, str(index_path))
            logger.debug('extract_transcriptions: {}'.format(t.delta))

            with Timer() as t:
                polygons = build_transcriptions(image, page, self.source, lines)

                for polygon in polygons:
                    queue.put(polygon)  # will block until a slot is free
            logger.debug('build_transcriptions: {}'.format(t.delta))

        logger.info('All files parsed')

    def save_transcriptions(self, queue):
        from queue import Empty
        while True:
            data = []
            try:
                while len(data) < settings.TRANSCRIPTIONS_IMPORT_CHUNK_SIZE:
                    data.append(queue.get(block=True, timeout=1))
            except Empty:
                pass

            if not data:
                continue

            with Timer() as t:
                transcriptions, _ = save_transcriptions(*data)
            logger.debug('save_transcriptions ({}): {}'.format(len(data), t.delta))
            with Timer() as t:
                index_transcriptions(transcriptions)
            logger.debug('index_transcriptions ({}): {}'.format(len(transcriptions), t.delta))

            # Mark all items as done
            for _ in range(len(data)):
                queue.task_done()

    def dry_run(self):
        for pages, index_path in self.list_index_pages():
            page = self.find_page(pages, index_path)
            if page:
                logger.info("{}\t{}".format(index_path, page.zone.image.path))
