from rest_framework import serializers
from arkindex.project.serializer_fields import PolygonField, EnumField, PointField, IIIFPathField
from arkindex.documents.models import Corpus
from arkindex.images.models import Image, Zone, ImageServer
from arkindex.dataimport.models import DataFile
from arkindex.project.aws import S3FileStatus
import re
import uuid


class ImageSerializer(serializers.ModelSerializer):
    """
    Serialises an image
    """
    status = EnumField(S3FileStatus)
    s3_url = serializers.SerializerMethodField()

    class Meta:
        model = Image
        fields = (
            'id',
            'path',
            'width',
            'height',
            'url',
            's3_url',
            'status',
        )
        read_only_fields = ('id', 'path', 'width', 'height', 'url', 's3_url', )

    def get_s3_url(self, obj):
        if 'request' not in self.context:
            return
        # Only allow the S3 URL for internal users or admins
        user = self.context['request'].user
        if user.is_authenticated and (user.is_admin or user.is_internal):
            return obj.s3_url

    def validate_status(self, value):
        if not self.instance:
            # Force the Unchecked status when creating a new image
            return S3FileStatus.Unchecked
        elif value == S3FileStatus.Checked:
            # Perform image validation if we are updating an existing image to Checked
            try:
                if not self.instance.server.read_only:
                    self.instance.check_hash(raise_exc=True)
                self.instance.perform_check(raise_exc=True)
            except (AssertionError, ValueError) as e:
                raise serializers.ValidationError(str(e))
        return value


class ImagePathDefault(object):
    """
    Default an image's path to its UUID.
    Required because the implicit UniqueTogetherValidator in the Image serializer requires
    both server and path to either be required or have a default value.
    """

    def set_context(self, serializer_field):
        self.serializer = serializer_field.parent

    def __call__(self):
        return str(self.serializer.fields['id'].default)


class ImageUploadSerializer(ImageSerializer):
    """
    Serialize an image to upload in s3
    """
    status = EnumField(S3FileStatus, read_only=True)
    hash = serializers.RegexField(re.compile(r'[0-9A-Fa-f]{32}'), min_length=32, max_length=32)
    datafile = serializers.PrimaryKeyRelatedField(
        queryset=DataFile.objects.none(),
        required=False,
        allow_null=True,
    )

    server = serializers.PrimaryKeyRelatedField(
        queryset=ImageServer.objects.exclude(s3_bucket=None),
        required=False,
        write_only=True,
    )
    path = IIIFPathField(
        required=False,
        default=ImagePathDefault(),
    )

    s3_put_url = serializers.SerializerMethodField()

    class Meta(ImageSerializer.Meta):
        fields = ImageSerializer.Meta.fields + ('hash', 'datafile', 's3_put_url', 'server')
        # Remove path, width and height from the ImageSerializer read-only fields
        read_only_fields = ('id', 'url', 's3_url', 's3_put_url')
        # Handle validation at API level
        validators = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # No request when generating OpenAPI schemas
            return
        self.fields['datafile'].queryset = DataFile.objects.filter(
            corpus__in=Corpus.objects.readable(self.context['request'].user),
        )
        self.fields['server'].default = ImageServer.objects.local

        # Set a fixed default for the image ID - allows the ID to be used as the image path
        self.fields['id'].default = uuid.uuid4()

    def get_s3_put_url(self, obj):
        if obj.status == S3FileStatus.Checked or not obj.server.is_local:
            # No PUT for existing images or external servers
            return None
        return obj.s3_put_url


class ZoneSerializer(serializers.ModelSerializer):
    polygon = PolygonField()
    center = PointField(source='polygon.center')
    image = ImageSerializer()

    class Meta:
        model = Zone
        fields = (
            'id',
            'polygon',
            'center',
            'url',
            'image',
        )
