from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.images.models import Image, ImageServer
from arkindex.users.models import User
from rest_framework import status
from django.test import override_settings
from django.urls import reverse
from unittest.mock import patch, call
import random


@override_settings(LOCAL_IMAGESERVER_ID=1)
class TestImageApi(FixtureAPITestCase):

    def setUp(self):
        # A random 32-character hex string
        self.image_hash = '{:032x}'.format(random.randrange(16**32))
        self.client.force_login(self.user)

    def test_create_image_requires_login(self):
        self.client.logout()
        response = self.client.post(reverse('api:image-create'), {'hash': self.image_hash})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    def test_create_image(self, s3_presigned_url_mock):
        """
        Use serializer to create a new Image
        Endpoint should return a s3 URI to upload image file
        """
        s3_presigned_url_mock.return_value = 'http://s3/upload_put_url'

        response = self.client.post(reverse('api:image-create'), {'hash': self.image_hash})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)
        im = Image.objects.get(id=data['id'])
        self.assertDictEqual(
            response.json(),
            {
                'id': str(im.id),
                'path': im.path,
                'width': im.width,
                'height': im.height,
                'url': im.url,
                'status': im.status.value,
                'hash': im.hash,
                's3_put_url': im.s3_put_url,
                's3_url': None,
                'datafile': None,
            }
        )
        self.assertEqual(im.hash, self.image_hash)
        self.assertEqual(im.s3_put_url, 'http://s3/upload_put_url')

    def test_create_image_unique_hash(self):
        response = self.client.post(reverse('api:image-create'), {'hash': self.image_hash})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        image_id = data['id']

        response = self.client.post(reverse('api:image-create'), {'hash': self.image_hash})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertDictEqual(data, {
            'hash': ['Image with this hash already exists'],
            'id': image_id,
        })

        self.assertEqual(Image.objects.filter(hash=self.image_hash).count(), 1)

    def test_create_image_unique_path(self):
        response = self.client.post(reverse('api:image-create'), {'hash': self.image_hash, 'path': 'something'})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        image_id = data['id']

        new_hash = '{:032x}'.format(random.randrange(16**32))
        response = self.client.post(reverse('api:image-create'), {'hash': new_hash, 'path': 'something'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertDictEqual(data, {
            'path': ['Image with this path already exists'],
            'id': image_id,
        })

        self.assertEqual(Image.objects.filter(path='something').count(), 1)

    def test_create_image_datafile(self):
        df = self.corpus.files.create(
            name='test.jpg',
            size=42,
            hash='aaaa',
            content_type='image/jpeg',
        )

        response = self.client.post(reverse('api:image-create'), {
            'hash': self.image_hash,
            'datafile': str(df.id),
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        img = Image.objects.get(id=data['id'])
        self.assertEqual(img.datafile, df)

    def test_create_image_custom_server(self):
        server = ImageServer.objects.create(
            display_name='S3 test',
            s3_bucket='buckette',
            url='http://some.server/',
        )
        response = self.client.post(reverse('api:image-create'), {
            'hash': self.image_hash,
            'server': server.id,
            'path': 'a/b/c',
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)
        image = Image.objects.get(id=data['id'])
        self.assertEqual(image.server_id, server.id)
        self.assertEqual(image.path, 'a%2Fb%2Fc')
        self.assertEqual(image.s3_bucket, 'buckette')

    def test_create_image_server_s3_only(self):
        server = ImageServer.objects.create(
            display_name='S3 test',
            s3_bucket=None,
            url='http://some.server/',
        )
        response = self.client.post(reverse('api:image-create'), {
            'hash': self.image_hash,
            'server': server.id,
            'path': 'a/b/c',
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_image_size_checked(self):
        """
        Test the CreateImage endpoint sets the image as checked when the
        request is made by an internal user and includes the image's size
        """
        internal_user = User.objects.create_internal_user('internal@test.com')
        internal_user.verified_email = True
        internal_user.save()
        self.assertTrue(internal_user.is_internal)
        self.client.force_login(internal_user)

        response = self.client.post(reverse('api:image-create'), {
            'hash': self.image_hash,
            'width': 1234,
            'height': 5678,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        image = Image.objects.get(id=data['id'])
        self.assertEqual(image.width, 1234)
        self.assertEqual(image.height, 5678)
        self.assertEqual(image.status, S3FileStatus.Checked)

    def test_create_image_checked_internal(self):
        """
        Test the CreateImage endpoint only sets the image as checked
        when the requester is an internal user
        """
        self.assertFalse(self.user.is_internal)
        response = self.client.post(reverse('api:image-create'), {
            'hash': self.image_hash,
            'width': 1234,
            'height': 5678,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        image = Image.objects.get(id=data['id'])
        # The image has a size, but is not checked
        self.assertEqual(image.width, 1234)
        self.assertEqual(image.height, 5678)
        self.assertEqual(image.status, S3FileStatus.Unchecked)

    def test_update_image_requires_login(self):
        self.client.logout()
        img = ImageServer.objects.local.images.create(
            status=S3FileStatus.Unchecked,
            path='test_update',
        )
        response = self.client.put(
            reverse('api:image-retrieve', kwargs={'pk': str(img.id)}),
            {'status': 'checked'},
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.images.serializers.Image.perform_check')
    @patch('arkindex.images.serializers.Image.check_hash')
    def test_update_image_check(self, check_hash_mock, perform_check_mock):
        """
        Test setting an image's status to Checked runs the image checks
        """
        img = ImageServer.objects.local.images.create(
            status=S3FileStatus.Unchecked,
            path='test_update',
        )
        response = self.client.put(
            reverse('api:image-retrieve', kwargs={'pk': str(img.id)}),
            {'status': 'checked'},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        img.refresh_from_db()
        self.assertEqual(img.status, S3FileStatus.Checked)
        self.assertEqual(check_hash_mock.call_count, 1)
        self.assertEqual(perform_check_mock.call_count, 1)
        self.assertEqual(check_hash_mock.call_args, call(raise_exc=True))
        self.assertEqual(perform_check_mock.call_args, call(raise_exc=True))

    @patch('arkindex.images.serializers.Image.perform_check')
    @patch('arkindex.images.serializers.Image.check_hash')
    def test_update_image_server_read_only(self, check_hash_mock, perform_check_mock):
        img = Image.objects.create(
            status=S3FileStatus.Unchecked,
            path='test_update',
            server=ImageServer.objects.create(
                display_name='S3 test',
                s3_bucket='s3_blanket',
                read_only=True,
                url='http://some.server/',
            )
        )
        response = self.client.put(
            reverse('api:image-retrieve', kwargs={'pk': str(img.id)}),
            {'status': 'checked'},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        img.refresh_from_db()
        self.assertEqual(img.status, S3FileStatus.Checked)
        self.assertEqual(check_hash_mock.call_count, 0)
        self.assertEqual(perform_check_mock.call_count, 1)
        self.assertEqual(perform_check_mock.call_args, call(raise_exc=True))
