from django.core.checks import register, Warning, Error
from ponos.recipe import parse_recipe
import os.path
import subprocess
import yaml


def only_runserver(func):
    "Decorator to run a system check only when running a dev server or deploying"
    from django.conf import settings

    def wrapper(*args, **kwargs):
        if settings.ALL_CHECKS:
            return func(*args, **kwargs)
        return []

    return wrapper


@register()
def api_urls_check(*args, **kwargs):
    '''
    Check that all urls end with a mandatory slash
    '''
    from arkindex.project.api_v1 import api
    return [
        Error(
            'This api url misses a mandatory ending /',
            hint=str(url),
            id='arkindex.E001',
        )
        for url in api
        if not str(url.pattern).endswith('/')
    ]


@register()
@only_runserver
def ml_tools_check(*args, **kwargs):
    """
    Check that the ML classifiers defined in settings actually exist
    """
    from django.conf import settings
    if not os.path.isdir(settings.ML_CLASSIFIERS_DIR):
        return [Error(
            'ML tools directory does not exist',
            hint='settings.ML_CLASSIFIERS_DIR = "{}"'.format(settings.ML_CLASSIFIERS_DIR),
            id='arkindex.E008',
        )]
    return []


@register()
@only_runserver
def local_imageserver_check(*args, **kwargs):
    """
    Check that the configured local ImageServer exists
    """
    from django.conf import settings
    from arkindex.images.models import ImageServer
    errors = []
    local_id = settings.LOCAL_IMAGESERVER_ID

    try:
        ImageServer.objects.get(id=local_id)
    except ImageServer.DoesNotExist:
        errors.append(Error(
            'Local ImageServer with ID {} does not exist'.format(local_id),
            hint='settings.LOCAL_IMAGESERVER_ID = {}'.format(local_id),
            id='arkindex.E004',
        ))

    return errors


@register()
def docker_images_check(*args, **kwargs):
    """
    Check that the Arkindex backend and ML images exist
    """
    from django.conf import settings
    errors = []

    if settings.PONOS_RECIPE is None:
        # In a Ponos task
        return []

    images = (
        (settings.ARKINDEX_APP_IMAGE, 'ARKINDEX_APP_IMAGE'),
        (settings.ARKINDEX_TASKS_IMAGE, 'ARKINDEX_TASKS_IMAGE'),
    )
    for image_tag, setting_name in images:
        try:
            subprocess.run(
                ['docker', 'image', 'inspect', image_tag],
                check=True,
                # Silent output
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError:
            errors.append(Error(
                'Docker image with tag "{}" was not found.'.format(image_tag),
                hint='settings.{} = "{}"'.format(setting_name, image_tag),
                id='arkindex.E006',
            ))
        except FileNotFoundError:
            # Docker is not available, ignore check
            pass

    return errors


@register()
def ponos_recipe_check(*args, **kwargs):
    """
    Check that the base Ponos recipe seems legit
    """
    from django.conf import settings
    errors = []

    if settings.PONOS_RECIPE is None:
        # In a Ponos task
        return []

    recipe = settings.PONOS_RECIPE.copy()
    # Add a dummy task because Ponos wants at least one task
    recipe['tasks'] = {'task1': {'image': 'hello-world'}}
    try:
        parse_recipe(yaml.dump(recipe))
    except (ValueError, AssertionError) as e:
        errors.append(Error(
            'Base Ponos recipe validation failed: {}'.format(str(e)),
            hint='settings.PONOS_RECIPE',
            id='arkindex.E007',
        ))

    return errors


@register()
@only_runserver
def internal_group_check(*args, **kwargs):
    from django.conf import settings
    from django.contrib.auth.models import Group
    try:
        group_id = settings.INTERNAL_GROUP_ID
    except AttributeError:
        return [Warning(
            'Internal group ID is not set; internal API requests will fail.',
            hint='settings.INTERNAL_GROUP_ID',
            id='arkindex.W001',
        )]
    try:
        Group.objects.get(id=group_id)
    except Group.DoesNotExist:
        return [Warning(
            'Internal user group does not exist; internal API requests will fail.',
            hint='settings.INTERNAL_GROUP_ID = "{}"'.format(group_id),
            id='arkindex.W002',
        )]
    return []


@register()
@only_runserver
def gitlab_oauth_check(*args, **kwargs):
    from django.conf import settings
    warnings = []
    app_id = getattr(settings, 'GITLAB_APP_ID', None)
    app_secret = getattr(settings, 'GITLAB_APP_SECRET', None)
    if not app_id:
        warnings.append(Warning(
            'GitLab app ID is not set; Git imports will fail.',
            hint='settings.GITLAB_APP_ID = {}'.format(repr(app_id)),
            id='arkindex.W003',
        ))
    if not app_secret:
        warnings.append(Warning(
            'GitLab app secret is not set; Git imports will fail.',
            hint='settings.GITLAB_APP_SECRET = {}'.format(repr(app_secret)),
            id='arkindex.W004',
        ))
    return warnings


@register()
@only_runserver
def s3_check(*args, **kwargs):
    from django.conf import settings
    aws_settings = {
        'AWS_ACCESS_KEY': 'AWS access key ID',
        'AWS_SECRET_KEY': 'AWS secret key',
        'AWS_ENDPOINT': 'AWS endpoint',
        'AWS_THUMBNAIL_BUCKET': 'S3 thumbnails bucket name',
        'AWS_STAGING_BUCKET': 'S3 staging bucket name',
    }
    errors = []
    for name, display_name in aws_settings.items():
        value = getattr(settings, name, None)
        if not value:
            errors.append(Error(
                '{} is missing; all S3-related features will fail.'.format(display_name),
                hint='settings.{} = {}'.format(name, repr(value)),
                id='arkindex.E011',
            ))
    return errors
