from unittest import TestCase
from django.test import RequestFactory
from rest_framework import serializers
from rest_framework.request import Request
from rest_framework.views import APIView
from rest_framework.schemas.openapi import SchemaGenerator
from arkindex.project.serializer_fields import EnumField
from arkindex.project.openapi import AutoSchema
from arkindex_common.enums import DataImportMode
import warnings


# Helper methods taken straight from the DRF test suite
def create_request(path):
    factory = RequestFactory()
    request = Request(factory.get(path))
    return request


def create_view(view_cls, method, request):
    generator = SchemaGenerator()
    view = generator.create_view(view_cls.as_view(), method, request)
    return view


class TestOpenAPI(TestCase):

    def test_no_deprecated(self):
        """
        Test deprecated is omitted in schemas if the attribute is absent
        """

        class ThingView(APIView):
            action = 'Retrieve'

            def get(self, *args, **kwargs):
                pass

        inspector = AutoSchema()
        inspector.view = create_view(ThingView, 'GET', create_request('/test/'))
        self.assertDictEqual(
            inspector.get_operation('/test/', 'GET'),
            {
                'operationId': 'RetrieveThing',
                'parameters': [],
                'responses': {
                  '200': {
                    'content': {
                        'application/json': {
                            'schema': {}
                        }
                    },
                    'description': '',
                  }
                }
            }
        )

    def test_deprecated_attribute(self):
        """
        Test the optional `deprecated` attribute on views
        """

        class ThingView(APIView):
            action = 'Retrieve'
            deprecated = True

            def get(self, *args, **kwargs):
                pass

        inspector = AutoSchema()
        inspector.view = create_view(ThingView, 'GET', create_request('/test/'))
        self.assertDictEqual(
            inspector.get_operation('/test/', 'GET'),
            {
                'operationId': 'RetrieveThing',
                'parameters': [],
                'responses': {
                  '200': {
                    'content': {
                        'application/json': {
                            'schema': {}
                        }
                    },
                    'description': '',
                  }
                },
                'deprecated': True,
            }
        )

    def test_overrides(self):
        """
        Test the optional `openapi_overrides` attribute on views
        """

        class ThingView(APIView):
            action = 'Retrieve'
            openapi_overrides = {
                'operationId': 'HaltAndCatchFire',
                'tags': ['bad-ideas'],
                'parameters': [
                    {
                        'description': 'Some extra parameter',
                        'in': 'query',
                        'name': 'something',
                        'required': False,
                        'schema': {'type': 'integer'},
                    }
                ],
            }

            def get(self, *args, **kwargs):
                pass

        inspector = AutoSchema()
        inspector.view = create_view(ThingView, 'GET', create_request('/test/{id}/'))
        self.assertDictEqual(
            inspector.get_operation('/test/{id}/', 'GET'),
            {
                'operationId': 'HaltAndCatchFire',
                'parameters': [
                    {
                        'description': '',
                        'in': 'path',
                        'name': 'id',
                        'required': True,
                        'schema': {'type': 'string'},
                    },
                    {
                        'description': 'Some extra parameter',
                        'in': 'query',
                        'name': 'something',
                        'required': False,
                        'schema': {'type': 'integer'},
                    }
                ],
                'responses': {
                  '200': {
                    'content': {
                        'application/json': {
                            'schema': {}
                        }
                    },
                    'description': '',
                  }
                },
                'tags': ['bad-ideas'],
            }
        )

    def test_bugfix_list_uppercase(self):
        """
        Test list API views have title-cased endpoint names
        """

        class ThingView(APIView):
            action = 'list'

            def get(self, *args, **kwargs):
                pass

        inspector = AutoSchema()
        inspector.view = create_view(ThingView, 'GET', create_request('/test/'))
        self.assertDictEqual(
            inspector.get_operation('/test/', 'GET'),
            {
                'operationId': 'ListThings',
                'parameters': [],
                'responses': {
                  '200': {
                    'content': {
                        'application/json': {
                            'schema': {
                                'type': 'array',
                                'items': {},
                            }
                        }
                    },
                    'description': '',
                  }
                },
            }
        )

    def test_bugfix_hstorefield(self):
        """
        Test the temporary fix for serializers.HStoreField being treated as a string
        """
        inspector = AutoSchema()
        self.assertDictEqual(
            inspector._map_field(serializers.HStoreField()),
            {'type': 'object'},
        )
        # Other fields should be unaffected
        self.assertDictEqual(
            inspector._map_field(serializers.CharField()),
            {'type': 'string'},
        )

    def test_bugfix_callable_defaults(self):
        """
        Test the temporary fix to convert callable defaults into their results
        """

        def bad_default():
            raise Exception('Nope')

        def fancy_default():
            raise Exception("Don't touch me!")

        fancy_default.openapi_value = 1337.0

        class ThingSerializer(serializers.Serializer):
            my_field = serializers.FloatField(default=float)
            my_field_on_fire = serializers.FloatField(default=bad_default)
            my_fancy_field = serializers.FloatField(default=fancy_default)

        inspector = AutoSchema()
        with warnings.catch_warnings(record=True) as warn_list:
            schema = inspector._map_serializer(ThingSerializer())

        self.assertEqual(len(warn_list), 1)
        self.assertEqual(
            str(warn_list[0].message),
            'Unsupported callable default for field my_field_on_fire: Nope',
        )

        self.assertDictEqual(schema, {
            'properties': {
                'my_field': {
                    'type': 'number',
                    'default': 0.0,
                },
                'my_field_on_fire': {
                    'type': 'number',
                },
                'my_fancy_field': {
                    'type': 'number',
                    'default': 1337.0,
                },
            },
        })

    def test_bugfix_enum_defaults(self):
        """
        Test the temporary fix to convert enums into strings
        """

        class ThingSerializer(serializers.Serializer):
            my_field = EnumField(DataImportMode, default=DataImportMode.Images)

        inspector = AutoSchema()
        self.assertDictEqual(inspector._map_serializer(ThingSerializer()), {
            'properties': {
                'my_field': {
                    'default': 'images',
                    'enum': [
                        'images',
                        'pdf',
                        'repository',
                        'elements',
                    ]
                }
            }
        })

    def test_bugfix_null_defaults(self):
        """
        Test the temporary fix that hides None defaults when they do not actually exist
        """

        class ThingSerializer(serializers.Serializer):
            my_field = serializers.CharField()

        inspector = AutoSchema()
        self.assertDictEqual(inspector._map_serializer(ThingSerializer()), {
            'properties': {
                'my_field': {
                    'type': 'string',
                }
            },
            'required': ['my_field'],
        })
